# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
"""

from importlib_metadata import entry_points
import numpy as np
from collections import Counter
import matplotlib.pyplot as plt
import math
plt.rcParams.update({'font.family': 'SimHei', 'mathtext.fontset': 'stix'})

###########################################################################################
###############################     1 Radar plot       ###############################
###########################################################################################

####***************************     1.1 Normalization       ****************************###

def radar_plot(df,plot_package='matplotlib',savefig=False,save_path='radar.png'):
    """ Draw a radar plot based on the given data with the type of pandas DataFrame

    Parameters
    ----------
    df           : a pandas object that contains specific columns

    plot_package : type of plotting package
        - 'matplotlib'(default)
        - 'plotly'

    savefig      : whether save the figure

    save_path    : where to save the figure

        
    Yields
    ------
    fig          : the radar plot

    """
    plt.rcParams['font.sans-serif'] = ['SimHei']
    plt.rcParams['axes.unicode_minus'] = False
    
    names = list(df.index)
    labels = list(df.columns)
    data = [dict(df.loc[names[i],:]) for i in range(len(names))]
    data_length = len(data[0])
    # 将极坐标根据数据长度进行等分
    angles = np.linspace(0, 2*np.pi, data_length, endpoint=False)

    score = [[v for v in d.values()] for d in data]

    # 设置图形的大小
    fig = plt.figure(figsize=(8, 6), dpi=100)
    # 新建一个子图
    ax = plt.subplot(111, polar=True)
    angles = np.concatenate((angles, [angles[0]]))
    labels = np.concatenate((labels, [labels[0]])) 
    for i in range(len(names)):
        score1 = np.concatenate((score[i],[score[i][0]]))
        # 绘制雷达图
        ax.plot(angles, score1)
    ax.set_rlim(df.values.min(), df.values.max())   
    # 设置雷达图中每一项的标签显示
    ax.set_thetagrids(angles*180/np.pi, labels)
    plt.legend(names)
    #plt.show()
    
    return fig






# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
"""

from importlib_metadata import entry_points
import numpy as np
from collections import Counter
import matplotlib.pyplot as plt
import math
from scipy.optimize import curve_fit
from scipy.integrate import odeint
from scipy.optimize import minimize

import warnings
warnings.filterwarnings('ignore')

plt.rcParams.update({'font.family': 'SimHei', 'mathtext.fontset': 'stix'})

###########################################################################################
###############################   1 Differential Equations  ###############################
###########################################################################################

####***************************     1.1 Population          ****************************###


def pop_estimate_coef(population,year,function_type='exponential'):
    """Calculate the coeficient(s) of a given type of growth model

    Parameters
    ----------
    population : a array or list type of data
    year            : year list.The default year list is range(len(population_list))
    function_type   : the type of grown mode
        - 'exp' : the exponential growth mode and its function is $x(t) = x_0 * e^{rt}$
        - 'logistic'    : the logistic growth mode and its funciton si $x(t) = \frac{x_m}{1+(\frac{x_m}{x_0})*e^{-rt}}$
        
    Yields
    ------
    coefs : the dict of coeficients in the given function
        - 'r'  : the nature growth rate
        - 'x_0'  : the initial population
        = 'x_m': the environmental carrying capacity

    """

    # convert population_list and year list to array type    
    population = np.array(population).flatten()
    year = np.array(year)    
    
    def logis_func(t,r,x0,xm):
        a1 = xm/x0 - 1
        a2 = np.e**(-r*(t-year[0]))
        return xm/(1+a1*a2)
    
    def exp_func(t,r,x0):
        return x0*np.e**(r*(t-year[0]))
    
    coefs = {'r':np.nan,'x_0':np.nan,'x_m':np.nan,'model_expression':np.nan}
    
    if function_type.lower() == 'exp':
        pFit,pCov = curve_fit(exp_func,year,population)
        r = pFit[0]
        x0 = pFit[1]
        coefs['r'] = r
        coefs['x_0'] = x0
        coefs['model_expression'] = f'{x0}*e^{r}x'
        print('The nature growth rate is ',r)
        print('The initial population is ',x0)
    elif function_type.lower() == 'logistic':
        pFit,pCov = curve_fit(logis_func,year,population,p0=np.ones(3))
        r = pFit[0]
        x0 = pFit[1]
        xm = pFit[2]
        coefs['r'] = r
        coefs['x_0'] = x0
        coefs['x_m'] = xm
        coefs['model_expression'] = f'{xm}/(1+{xm/x0}*e^{r}x'
        print('The nature growth rate is ',r)
        print('The initial population is ',x0)
        print('The environmental carrying capacity is ',xm)
    else:
        print("Invalid model name input. \n This function provides only exponential models ('exp') and logistic models ('logistic)'")

    return coefs
    


####***************************     1.1 Infectious Diseases  Simulation        ****************************###

def disease_simulate(total_steps,delta=1,start_time=0,lam=0.01,mu=0.01,sigma=0,i_ratio=0.01,r_ratio=0.0,e_ratio=0,model_type='sir'):
    """Simulate the future number of different divisions of people including suspective, infected and recovered people

    Parameters
    ----------
    total_steps : total steps for the simulation
    delta       : step length
    lam         : average number of people infected by a infected patient
    mu          : recovery rate
    i_ratio     : initial number of infected people
    r_ratio     : initial number of recovered or immune people
    e_ratio     : initial number of  people in incubation period
    model_type  : type of infectious disease model you want to choose
         - 'exp': Exponential model
         - 'si'  : SI model
         - 'sis' : SIS model
         - 'sir' : SIR model
         = 'seir': SEIR model
        
    Yields
    ------
    results : the dict of coeficients in the given model
        - 's_list'  : a ratio list of historic number of suspective people  
        - 'i_list'  : a ratio list of historic number of infective  people
        - 'r_list'  : a ratio list of historic number of recovered  people
        - 'e_list'  : a ratio list of historic number of people in incubation/latent period
    """

    
    results = {'time_list':np.nan,'s_list':np.nan,'i_list':np.nan,'r_list':np.nan,'e_list':np.nan}
    time_list = [start_time]
    s_list = [1-i_ratio-r_ratio-e_ratio]
    i_list = [i_ratio]
    r_list = [r_ratio]
    e_list = [e_ratio]

    # Exponential Model
    def exp():
        # Solve the difference equation recursively
        for i in range(total_steps):
            i_new = i_list[-1] + lam * i_list[-1]
            s_new = 0
            r_new = 0
            e_new = 0

            i_list.append(i_new)
            s_list.append(s_new)
            r_list.append(r_new)  
            e_list.append(e_new) 
         
            t_new = time_list[-1] + delta * time_list[-1]
            time_list.append(t_new)
        return time_list,i_list,s_list,r_list, e_list 

    # SI Model
    def si():
        for i in range(total_steps):
            i_new = i_list[-1] + lam * i_list[-1] * s_list[-1]
            i_list.append(i_new)
            s_new = s_list[-1] - lam * i_list[-1] * s_list[-1]
            r_new = 0 
            e_new = 0 

            i_list.append(i_new)
            s_list.append(s_new)
            r_list.append(r_new)  
            e_list.append(e_new) 

            t_new = time_list[-1] + delta * time_list[-1]
            time_list.append(t_new)
        return time_list,i_list, s_list, r_list, e_list       

    # SIS Model
    def sis():
        for i in range(total_steps):
            i_new = i_list[-1] + (lam-mu) * i_list[-1] * s_list[-1]
            s_new = s_list[-1] - (lam-mu) * i_list[-1] * s_list[-1]
            r_new = 0
            e_new = 0

            i_list.append(i_new)
            s_list.append(s_new)
            r_list.append(r_new)  
            e_list.append(e_new)   

            t_new = time_list[-1] + delta * time_list[-1]
            time_list.append(t_new)
        return time_list,i_list, s_list, r_list, e_list       

    # SIR Model
    def sir():
        for i in range(total_steps):
            i_new = i_list[-1] + (lam-mu) * i_list[-1] * s_list[-1]
            s_new = s_list[-1] - lam * i_list[-1] * s_list[-1]
            r_new = r_list[-1] + mu * i_list[-1]
            e_new = 0

            i_list.append(i_new)
            s_list.append(s_new)
            r_list.append(r_new)  
            e_list.append(e_new) 
            
            t_new = time_list[-1] + delta * time_list[-1]
            time_list.append(t_new)
        return time_list,i_list, s_list, r_list , e_list
        
    # SEIR Model
    def seir():
        for i in range(total_steps):
            e_new = e_list[-1] + lam * i_list[-1] * s_list[-1] - sigma * e_list[-1]
            s_new = s_list[-1] - lam * i_list[-1] * s_list[-1]
            i_new = i_list[-1] + sigma * e_list[-1] - mu * i_list[-1]
            r_new = r_list[-1] + mu * i_list[-1]

            i_list.append(i_new)
            s_list.append(s_new)
            r_list.append(r_new)  
            e_list.append(e_new) 

            t_new = time_list[-1] + delta * time_list[-1]
            time_list.append(t_new)
        return time_list,i_list, s_list, r_list , e_list  

    if model_type.lower() == 'exp':
        results['time_list'],results['i_list'],results['s_list'],results['r_list'],results['e_list'] = exp()
    elif model_type.lower() == 'si':
        results['time_list'],results['i_list'],results['s_list'],results['r_list'],results['e_list']  = si()
    elif model_type.lower() == 'sis':
        results['time_list'],results['i_list'],results['s_list'],results['r_list'],results['e_list']  = sis()
    elif model_type.lower() == 'sir':
        results['time_list'],results['i_list'],results['s_list'],results['r_list'],results['e_list']  = sir()
    elif model_type.lower() == 'seir':
        results['time_list'],results['i_list'],results['s_list'],results['r_list'],results['e_list']  = seir()
    else:
        print("Invalid model name input. \n This function provides only :\n - exponential model ('exp') , \n - SI model ('si)' \n -SIS model ('sis) \n - SIR model('sir) \n - SIER model('seir)")
        results.clear()
    return results


    
####***************************     1.2 Estimate Coefficients       ****************************###

def disease_estimate_coef(pop_array,time_array,model_type='sir',metric='square_sum'):
    """Calculate the coeficient(s) of a given type of infectious disease models

    Parameters
    ----------
    pop_array : a two-dimensional array consisting of the number of historical infections, 
                    the number of susceptible, the number of recovered, and the number of latent patients
    time_array: a array of time
    model_type : the type of models
        - 'exp' : the exponential growth model; pop_array consist of the number of people infected over time
        - 'si'  : the SI model; pop_array consist of the number of infected and susceptible persons over time
        - 'sis' : the SIS model; pop_array consist of the number of infected and susceptible persons over time
        - 'sir' : the SIR model; pop_array consist of the number of infected, susceptible and recovered persons over time
        - 'seir': the SIR model; pop_array consist of the number of infected, susceptible, recovered and latent persons over time
    metric    :  the metirc to measure the goodness of fitting
        - 'square_sum' : sum of errors squared
        - 'abs_sum'    : sum of absolute errors


    Yields
    ------
    coefs : the dict of coeficients in the given function
        - 'lam'  : transmission rate (for Exponential, SI,SIS,SIR,SEIR)
        - 'mu'   : recovery rate (for SIR, SEIR)
        = 'gamma': turn positive rate (for SEIR)

    """
    # initialize the result dictionary
    coefs = {'lam':np.nan,'mu':np.nan,'gamma':np.nan}

    #--------------Exp-------------------#
    if model_type.lower()=='exp':
        def est_exp(x):
            lam = x
            def d_exp(y,t):
                i = y
                didt = lam*i
                return didt
            Y = odeint(d_exp,pop_array[0],time_array)
            if metric.lower() == 'square_sum':
                return ((Y-pop_array)**2).sum()
            else:
                return (np.abs(Y-pop_array)).sum()
        x0 = 1
        bounds = [(0,10)]
        res = minimize(est_exp,x0,bounds=bounds)
        coefs['lam'] = res['x'][0]

    #--------------SI-------------------#
    elif model_type.lower()=='si':
        def est_si(x):
            lam = x
            def d_si(y,t):
                i = y
                didt = lam*i*(1-i)
                return didt
            Y0 = pop_array[0,0]
            Y = odeint(d_si,Y0,time_array)
            if metric.lower() == 'square_sum':
                return ((Y-pop_array)**2).sum()
            else:
                return (np.abs(Y-pop_array)).sum()
        x0 = 1
        bounds = [(0,10)]
        res = minimize(est_si,x0,bounds=bounds)
        coefs['lam'] = res['x'][0]    
        
    #--------------SIS-------------------#
    elif model_type.lower()=='sis':
        def est_sis(x):
            lam,mu = x
            def d_sis(y,t):
                i,s = y
                didt = lam*i*s-mu*i
                dsdt = -lam*i*s+mu*i
                return [didt,dsdt]
            Y0 = [pop_array[0,0],pop_array[0,1]]
            Y = odeint(d_sis,Y0,time_array)
            if metric.lower() == 'square_sum':
                return ((Y-pop_array)**2).sum()
            else:
                return (np.abs(Y-pop_array)).sum()
        x0 = [1,1]
        bounds = [(0,10),(0,10)]
        res = minimize(est_sis,x0,bounds=bounds)
        coefs['lam'],coefs['mu'] = res['x'][0],res['x'][1]    

    #--------------SIR-------------------#
    elif model_type.lower()=='sir':
        def est_sir(x):
            lam,mu = x
            def d_sir(y,t):
                i,s,r = y
                didt = lam*i*s-mu*i
                dsdt = -lam*i*s
                drdt = mu*i
                return [didt,dsdt,drdt]
            Y0 = [pop_array[0,0],pop_array[0,1],pop_array[0,2]]
            Y = odeint(d_sir,Y0,time_array)
            if metric.lower() == 'square_sum':
                return ((Y-pop_array)**2).sum()
            else:
                return (np.abs(Y-pop_array)).sum()
        x0 = [1,1]
        bounds = [(0,10),(0,10)]
        res = minimize(est_sir,x0,bounds=bounds)
        coefs['lam'],coefs['mu'] = res['x'][0],res['x'][1]   

    #--------------SEIR-------------------#
    elif model_type.lower()=='seir':
        def est_seir(x):
            lam,mu,gamma = x
            def d_seir(y,t):
                i,s,r,e = y
                didt = gamma*e
                dedt = lam*i*s-gamma*e
                dsdt = -lam*i*s
                drdt = mu*i
                return [didt,dsdt,drdt,dedt]
            Y0 = [pop_array[0,0],pop_array[0,1],pop_array[0,2],pop_array[0,3]]
            Y = odeint(d_seir,Y0,time_array)
            if metric.lower() == 'square_sum':
                return ((Y-pop_array)**2).sum()
            else:
                return (np.abs(Y-pop_array)).sum()
        x0 = [1,1,1]
        bounds = [(0,10),(0,10),(0,10)]
        res = minimize(est_seir,x0,bounds=bounds)
        coefs['lam'],coefs['mu'],coefs['gamma'] = res['x'][0],res['x'][1],res['x'][2]         
    
    return coefs

        



# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
"""
import pandas as pd
from pulp import LpProblem, LpMinimize, LpVariable, lpSum, value
from datetime import datetime
import numpy as np

#################################################################################################
###############################     1 Heuristic Algorithms        ###############################
#################################################################################################

####***************************     1.1 Simulated Annealing       ****************************###

def Simulated_Annealing(f, x, alpha=.99, t=10000, delta=.1, maxIter=1000):
        
    ''' Simulated Annealing Algorithm (objective: find global minimum)

    Parameters:
    ----------
    f - objective function, R^n -> R
    x - inital solution, starting point, R^n
    alpha - annealing schedule parameter
    t - inital temperature
    delta - neighborhood radius
    maxIter - maximum no. of iterations  
    
    Yields :
    --------
    result : a dictionary contains the following parameters
        - 'x_opt' : the optimal x value(s)
        - 'f_opt' : the optimal value of the objective function
        - 'x_hist': the array of x values
        - 'f_hist': the array of the objective function
        - 'time'  : time
    ''' 

    
    # initializing starting parameters
    results = {'x_opt':x, 'f_opt':f(x), 'x_hist':[x], 'time':[0],
               'f_hist':[f(x)], 'temp':[t], 'transProb':[0]}
    
    currIter = 1
    finished = False
    x_s = x    
    time_0 = datetime.now() # to measure speed
    
    while not finished:
        
        # x_c - uniformly drawing a candidate solution from neighborhood of x_s
        unif = np.random.rand(len(x_s))
        x_c = x_s + (-delta + 2*delta*unif)
    
        # A - calculating Metropolis activation function
        A = np.minimum(1, np.exp(-(f(x_c) - f(x_s)) / t))
    
        # transition to candidate solution
        if bool(np.random.rand(1) < A):
            x_s = x_c
        
        # temperature update for the next iteration
        t = alpha * t

    
        if currIter < maxIter:
            
            # if better solution, update results
            if f(x_s) < f(results['x_opt']):
                results['x_opt'] = x_s
                results['f_opt'] = f(x_s)
            
            # update results history
            results['x_hist'].append(x_s)
            results['f_hist'].append(f(x_s))
            results['time'].append((datetime.now() - time_0).microseconds)

            results['transProb'].append(A)
        else:
            finished = True
        
        # if currIter % 250 == 0:
        #     print(f"f_opt after {currIter} iterations: {results['x_opt']} \n")
        
        currIter += 1
    print('The optimal x:',results['x_opt'])
    print('The optimal objective function :',results['f_opt'])
    return results['x_opt'],results['f_opt']


####***************************     1.2 Genetic Algorithm       ****************************###

# two helper functions
def binToInt(x):
    # Translate the binary chromosome to real values
    flipped = np.flipud(x)
    idx = np.argwhere(flipped==1).reshape(-1,)
    return (2**idx).sum()
  
def getCoords(population, cel, x_min, x_max):
    # Transform the binary chromosome of size 'cel' into real values of size 2
    coords = np.zeros((population.shape[0], 2))
    for i in range(population.shape[0]):
        for j in range(2): # test for more dimensions in spare time
            coordTemp = binToInt(population[i, (j*cel):((j+1)*cel)])
            # ensuring we are not leaving bounding box
            coords[i, j] = ((x_max[j]-x_min[j])/(2**cel))*coordTemp + x_min[j]
            
    return(coords)


def Genetic_Algorithm(f, x_min=[-20, -20], x_max=[20, 20], cel=50,
                     popSize=30, pMut=0.05, maxIter=1000):

    ''' Genetic Algorithm (objective: find global minimum)

    Parameters:
    ----------
    - f: objective function, R^n -> R
    - x_min: vector of the minimum values of coordinates, 
    - x_max: vector of the maximum values of coordinates
    - cel: coordinate encryption length, number of genes in a single chromosome
    - popSize: size of the population
    - pMut: probability of single genome mutation
    - maxIter: number of generations
    
    Yields :
    --------
    results : a dictionary contains the following parameters
        - 'x_opt' : the optimal x value(s)
        - 'f_opt' : the optimal value of the objective function
        - 'x_hist': the array of x values
        - 'f_hist': the array of the objective function
        - 'time'  : time
    ''' 

  
    # initializing history
    results = {'x_opt':[], 'f_opt':[], 'x_hist':[], 'f_mean':[], 
               'f_hist':[], 'time':[]}

    # Check the number of dimensions
    d = len(x_min) 
        
    # Initialize population
    population = np.zeros((popSize, cel*d))
      
    for i in range(popSize):
        # .5 chosen arbitrarily
        population[i,] = np.random.uniform(size=cel*d) > .5 
    
    coordinates = getCoords(population, cel, x_min, x_max)
      
    # Calculate fittness of individuals
    objFunction = np.zeros((popSize,))
    for i in range(popSize):
        objFunction[i] = f(coordinates[i,])
    
    # Assign the first population to output 
    results['x_opt'] = coordinates[np.argmin(objFunction),]
    results['f_opt'] = f(coordinates[np.argmin(objFunction),])
      
    # The generational loop
    finished = False
    currIter = 1
    time_0 = datetime.now() # to measure speed

    
    while not finished:
        # Assign the output
        if currIter <= maxIter:
            if results['f_opt'] > f(coordinates[np.argmin(objFunction),]):
                results['x_opt'] = coordinates[np.argmin(objFunction),]
                results['f_opt'] = f(coordinates[np.argmin(objFunction),])
          
            results['f_hist'].append(results['f_opt'])
            results['x_hist'].append(coordinates[np.argmin(objFunction),])
            results['f_mean'].append(np.mean(objFunction))
            results['time'].append((datetime.now() - time_0).microseconds)
        else:
          finished = True

        
        # Translate binary coding into real values to calculate function value
        coordinates = getCoords(population, cel, x_min, x_max)
        
        # Calculate fittness of the individuals
        objFunction = np.zeros((popSize,))
        for i in range(popSize):
            objFunction[i] = f(coordinates[i,])
        
        np.warnings.filterwarnings('ignore')

        rFitt = np.divide(min(objFunction), objFunction) # relative fittness
        # relative normalized fittness (sum up to 1) :
        nrFitt = np.divide(rFitt, sum(rFitt))
                
        # Selection operator (roulette wheel), analogy to disk
        selectedPool = np.zeros((popSize,))
        for i in range(popSize):
            selectedPool[i] = np.argmin(np.random.uniform(size=1) > np.cumsum(nrFitt))

        
        # Crossover operator (for selected pool)
        nextGeneration = np.zeros((popSize, cel*d))
        for i in range(popSize):
            parentId = int(np.round(np.random.uniform(1, popSize-1, 1)))
            cutId = int(np.round(np.random.uniform(1, d*cel-2, 1)))
            # Create offspring
            nextGeneration[i, :cutId] = population[int(selectedPool[i]), :cutId]
            nextGeneration[i, cutId:(d*cel)] = population[int(selectedPool[parentId]), cutId:(d*cel)]
        
        # Mutation operator
        for i in range(popSize):
            # Draw the genomes that will mutate
            genomeMutId = np.argwhere(np.random.rand(d*cel) < pMut)
            for j in range(len(genomeMutId)):
                nextGeneration[i, genomeMutId[j]] = not nextGeneration[i, genomeMutId[j]] 
        
        # Replace the old population
        population = nextGeneration
        currIter += 1

    print('The optimal x:',results['x_opt'])
    print('The optimal objective function :',results['f_opt'])
    return results['x_opt'],results['f_opt']


####***************************     1.3 Particle Swarm Optimization       ****************************###

def PSO(f, swarm_size=20, max_iter=200, x_min=[-20,-20], x_max=[20,20],
        c1=1, c2=1, omega=.5):

    ''' Particle Swarm Optimization (objective: find global minimum)

    Parameters:
    ----------
    - f: objective function, R^n -> R,
    - x_min: vector of the minimum values of coordinates, 
    - x_max: vector of the maximum values of coordinates,
    - swarm_size: number of particles in the swarm
    - max_iter: maximum number of iterations
    - c1: weight of personal best result of a particule
    - c2: weight of global best result of a swarm
    - omega: weight of current velocity
    
    Yields :
    --------
    results : a dictionary contains the following parameters
        - 'x_opt' : the optimal x value(s)
        - 'f_opt' : the optimal value of the objective function
        - 'x_hist': the array of x values
        - 'f_hist': the array of the objective function
        - 'time'  : time
    '''

    
    dim = len(x_min) 
    r = np.empty((swarm_size, dim))
    s = np.empty((swarm_size, dim))
    
    # initializing swarm
    swarm = np.empty((swarm_size, dim))
    velocity = np.empty((swarm_size, dim))
    p_best = np.empty((swarm_size, dim))
    swarm_result = np.empty((swarm_size, 1))
    
    # for each particle and each dimension initialize starting points
    for i in range(swarm_size):
        for j in range(dim):
            swarm[i, j] = np.random.uniform(low=x_min[j], high=x_max[j], size=1)
        velocity[i,:] = np.random.uniform(low=0, high=1, size=dim)
        p_best[i,:] = swarm[i,:]
        swarm_result[i] = f(swarm[i,:])
        g_best = swarm[np.argmin(swarm_result),:] # updating global best solution
    
    results = {'x_opt': [g_best], 'f_opt':[f(g_best)], 'x_hist':[g_best],
               'f_hist':[f(g_best)], 'time':[0]}

    
    pocz_iter = datetime.now()
    
    for k in range(max_iter):
        
        for m in range(swarm_size):
            r[m, :] = np.random.uniform(low=0, high=1, size=dim)
            s[m, :] = np.random.uniform(low=0, high=1, size=dim)
            
            # calculating components of new velocity
            old_vel = omega * velocity[m,:] # old velocity comp.
            best_pers_vel = c1 * r[m,:] * (p_best[m,:]-swarm[m,:]) # personal best comp.
            best_glob_vel = c2 * s[m,:] * (g_best-swarm[m,:]) # global best comp.
            # calculating new velocity
            velocity[m, :] = old_vel + best_pers_vel + best_glob_vel 
            # moving a particle in a new direction
            swarm[m, :] += velocity[m, :]
            
            # updating best solution for particle m
            if f(swarm[i,]) < f(p_best[i,]):
                p_best[m, :] = swarm[m, :]
            swarm_result[m] = f(swarm[m, :])
 

       
        # updating global best particle in iteration k
        if min(swarm_result)[0] < f(g_best):
            g_best = swarm[np.argmin(swarm_result), :]
        
        # saving history
        if results['f_opt'] > f(g_best):
            results['x_opt'] = swarm[np.argmin(swarm_result),:]
            results['f_opt'] = f(swarm[np.argmin(swarm_result),:])
        
        results['x_hist'].append(g_best)
        results['f_hist'].append(f(g_best))
        results['time'].append((datetime.now()-pocz_iter).microseconds)
    print('The optimal x:',results['x_opt'])
    print('The optimal objective function :',results['f_opt']) 
    return results['x_opt'],results['f_opt']


#################################################################################################
###############################     2 DEA                         ###############################
#################################################################################################

def dea(df_input,df_output,model_type='CRS'):
    ''' Data Evelopment Analysis(DEA) model

    Parameters:
    ----------
    - df_input : a DataFrame object containing the input data
    - df_output: a DataFrame object containing the output data
    - model_type:
        - 'CRS': CRS
        - 'VRS': VRS

    Yields :
    --------
    df_result : the DataFrame of the result

    '''    
    # Set building
    K = df_input.index
    I = df_input.columns
    J = df_output.columns
    X = df_input.to_dict()
    Y = df_output.to_dict()
    
    eta_plus = {}
    eta_minus={}
    objective_ = {}
    eps=0.00000001
    r = 0
    
    for k in K:
        # Model Building
        model = LpProblem(model_type+'_'+k, LpMinimize)  

        # Decision variables Building
        theta_r = LpVariable('theta_r',lowBound=0,upBound=1)
        lambda_k = LpVariable.dicts('lambda_k', lowBound=0, indexs=K)
        eta_plus_i = LpVariable.dicts('eta_plus_i',lowBound=0,indexs=I)
        eta_minus_j = LpVariable.dicts('eta_minus_j',lowBound=0,indexs=J)

        # Objective Function setting
        model += theta_r-eps*lpSum(eta_plus_i[i] for i in I)-eps*lpSum(eta_minus_j[j] for j in J)

        # Constraints setting
        for i in I:
            model += lpSum([
                    lambda_k[k] * X[i][k]
                for k in K]) + eta_plus_i[i] == theta_r * float(X[i][K[r]])
        for j in J:
            model += lpSum([
                    lambda_k[k] * Y[j][k]
                for k in K]) - eta_minus_j[j] == float(Y[j][K[r]])
            
        if model_type.upper()=='VRS':
            model += lpSum([ lambda_k[k] for k in K]) == 1
        # Model solving
        model.solve()
        
        eta_plus[k] = {}
        eta_minus[k] = {}
        for i in I:
            eta_plus[k][i] = value(eta_plus_i[i])
        for j in J:
            eta_minus[k][j] = value(eta_minus_j[j])
        objective_[k] = value(model.objective)
        
        r += 1
    df_obj = pd.DataFrame(pd.Series(objective_),columns=['Efficency']).T
    df_plus = pd.DataFrame(eta_plus)
    df_minus = pd.DataFrame(eta_minus)
    df_result = pd.concat([df_obj,df_plus,df_minus]).T
    return df_result



# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
Refer to 
1. https://github.com/LinBaiTao/Supplier_Selection_Methods 
2.https://github.com/akestoridis/mcdm
"""

from importlib_metadata import entry_points
import pandas as pd
import numpy as np
from collections import Counter
import matplotlib.pyplot as plt
plt.rcParams.update({'font.family': 'SimHei', 'mathtext.fontset': 'stix'})

###########################################################################################
###############################     1 Preprocessing         ###############################
###########################################################################################

####***************************     1.1 Normalization       ****************************###

def Normalize_Weights(weights_array,norm_type = 'divide_by_sum'):
    """Normalizes a provided weight array so that the sum equals 1

    Parameters
    ----------
    weights_array : a numpy array containing the raw weights
    
    norm_type : a string specifying the type of normalization to perform
        - 'divide_by_max' divides all values by the maximum value
        - 'divide_by_sum' divides all values by the sum of the values

    Yields
    ------
    temp_weights_array: a copy of the passed weights array with the normalizations performed

    Examples
    --------
    >>> import numpy as np
    >>> import mcdm_functions as mcfunc
    >>> criteria_weights = np.array([2,4,6,7,9])
    >>> temp = mcfunc.Normalize_Weights(criteria_weights,'divide_by_max')
    >>> print(temp)
    
        [ 0.22222222  0.44444444  0.66666667  0.77777778  1.        ]
    """   
    
    temp_weights_array = weights_array.copy()
    
    if norm_type is 'divide_by_max':
        temp_weights_array = temp_weights_array/temp_weights_array.max()

    elif norm_type is 'divide_by_sum':
        temp_weights_array = temp_weights_array/temp_weights_array.sum()
        
    else:
        print('You did not enter a valid type, so no changes were made')
    
    return temp_weights_array


###########################################################################################
def Normalize_Column_Scores(df, columns, norm_type = 'divide_by_max'):
    """Normalizes scores for specified columns in a pandas dataframe

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns
    
    columns: a list object that includes the columns to normalize
    
    norm_type : a string specifying the type of normalization to perform
        - 'divide_by_max' divides all values by the maximum value
        - 'range_norm' divides all values (+ the min) by the range of values in the column
        - 'z_norm' computes a z-score based on the mean and standard deviation of values
        - 'divide_by_sum' divides all values by the sum of the values
        - 'vector' dives all values by the square root of the sum of the squares of all values

    Yields
    ------
    temp_df: a copy of the passed dataframe with the normalizations performed

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> import mcdm_functions as mcfunc

    >>> data_dict = {'Product': ['A', 'B', 'C', 'D'],
                 'Product Advantage': [13.1,13.2,12.2,13.2],
                 'Strategic Alignment': [9.8,8.2,10.0,9.6],
                 'Technical Feasibility': [20.0,18.7,18.5,17.1],
                 'Market Attractiveness': [15.5,12.3,13.1,13.1]}
    >>> score_data = pd.DataFrame(data_dict)
    >>> score_data = score_data.set_index('Product')
    >>> print(score_data)
    
            Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                         15.5               13.1                  9.8   
    B                         12.3               13.2                  8.2   
    C                         13.1               12.2                 10.0   
    D                         13.1               13.2                  9.6   

             Technical Feasibility  
    Product                         
    A                         20.0  
    B                         18.7  
    C                         18.5  
    D                         17.1  
    
    
    >>> columns = ['Market Attractiveness','Product Advantage']
    >>> temp = mcfunc.Normalize_Column_Scores(score_data,columns)
    >>> print(temp)
    
             Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                     1.000000               13.1                  9.8   
    B                     0.793548               13.2                  8.2   
    C                     0.845161               12.2                 10.0   
    D                     0.845161               13.2                  9.6   

             Technical Feasibility  
    Product                         
    A                         20.0  
    B                         18.7  
    C                         18.5  
    D                         17.1  
    """   
    
    temp_df = df.copy()
    
    for column in columns:
        if norm_type is 'divide_by_max':
            max_entry = temp_df[column].max()
            temp_df[column] = temp_df[column]/max_entry
        
        elif norm_type is 'range_norm':
            min_entry = temp_df[column].min()
            max_entry = temp_df[column].max()
            temp_df[column] = (temp_df[column]-min_entry)/(max_entry - min_entry)
        
        elif norm_type is 'z_norm':
            mean = temp_df[column].mean()
            sd = temp_df[column].std()
            temp_df[column] = (temp_df[column]-mean)/sd
        
        elif norm_type is 'divide_by_sum':
            temp_df[column] = temp_df[column]/temp_df[column].sum()
            
        elif norm_type is 'vector':
            values = temp_df[column].values
            values_squared = values**2
            vector_norm = values/np.sqrt(np.sum(values_squared))
            temp_df[column] = vector_norm
        
        else:
            print('You did not enter a valid type, so no changes were made')
        
    return temp_df


####***************************     1.2 Conversion       ****************************###
def Convert_Higher_is_Better(df, columns, conversion_type = 'absolute'):
    """Converts scores given in a "lower is better" format to a "higher is better" format

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns
    
    columns: a list object that includes the columns to normalize
    
    conversion_type : a string specifying the type of conversion to perform
        - 'absolute' converts based on absolute scale
        - 'relative' converts based on relative scale
        - 'inverse'  converts based on its inverse value
        
    Yields
    ------
    temp_df: a copy of the passed dataframe with the conversions performed
    """
    
    
    temp_df = df.copy()
        
    for column in columns:
        if conversion_type.lower() is 'absolute':
            new_column = column+' (absolute HIB)'
            max_entry = temp_df[column].max()
            temp_df[new_column] = max_entry - temp_df[column]
    
        elif conversion_type.lower() is 'relative':
            new_column = column+' (relative HIB)'
            min_entry = temp_df[column].min()
            max_entry = temp_df[column].max()
            temp_df[new_column] = (max_entry - temp_df[column])/(max_entry-min_entry)
        
        elif conversion_type.lower() is 'inverse':
            new_column = column+' (inverse HIB)'
            temp_df[new_column] = 1/temp_df[column]

        else:
            print('You did not enter a valid type, so no changes were made')        
        
    return temp_df    


###########################################################################################
###############################     2 Weighting             ###############################
###########################################################################################

####***************************     2.1 Correlation         ****************************###

def abspearson(z_matrix):
    """ Calculate the absolute value of the Pearson correlation of the previded matrix.

    Parameters
    ----------
    z_matrix : a pandas or array object that contains specific columns

    
    Yields:
    -------
    abs_corr : a matrix of the Pearson correlation coefficients

    """
    
    # Make sure that the provided matrix is a float64 NumPy array
    z_matrix = np.array(z_matrix, dtype=np.float64)

    abs_corr = np.absolute(np.corrcoef(z_matrix, rowvar=False))

    return abs_corr


def pearson(z_matrix):
    """ Calculate the Pearson correlation coefficient of the provided matrix

    Parameters :
    ----------
    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    p_corr : the Pearson correlation coefficients of the provided matrix.
    """
    # Make sure that the provided matrix is a float64 NumPy array
    z_matrix = np.array(z_matrix, dtype=np.float64)

    p_corr = np.corrcoef(z_matrix, rowvar=False)
    return p_corr


def dcor(z_matrix):
    """ Calculate the distance correlation coefficients of the provided matrix
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    dcor_matrix : the distance correlation matrix

    """
    # Make sure that the provided matrix is a float64 NumPy array
    z_matrix = np.array(z_matrix, dtype=np.float64)

    # Initialize the matrix for the distance correlation coefficients
    dcor_matrix = np.ones(
        (z_matrix.shape[1], z_matrix.shape[1]),
        dtype=np.float64,
    )

    # Compute the matrix of squared distance covariances
    dcov2_matrix = squared_dcov_matrix(z_matrix)

    # Compute the distance correlation coefficients
    for j_col in range(z_matrix.shape[1]):
        # Get the squared distance variance of the j-th criterion
        j_dvar2 = dcov2_matrix[j_col, j_col]

        for l_col in range(j_col + 1, z_matrix.shape[1]):
            # Get the squared distance variance of the l-th criterion
            l_dvar2 = dcov2_matrix[l_col, l_col]

            # Compare the product of their squared distance variances
            if j_dvar2 * l_dvar2 == 0.0:
                # The two criteria are independent
                dcor_matrix[j_col, l_col] = 0.0
                dcor_matrix[l_col, j_col] = 0.0
            else:
                # Get the squared distance covariance of the two criteria
                jl_dcov2 = dcov2_matrix[j_col, l_col]

                # Compute the squared distance correlation of the two criteria
                jl_dcor2 = squared_dcor(jl_dcov2, j_dvar2, l_dvar2)

                # Compute the distance correlation of the two criteria
                dcor_matrix[j_col, l_col] = np.sqrt(jl_dcor2)
                dcor_matrix[l_col, j_col] = dcor_matrix[j_col, l_col]

    return dcor_matrix


def squared_dcov_matrix(z_matrix):
    """ Calculate the matrix of squared distance covariance between the columns of
    the provided matrix.
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    dcor2_matrix : the matrix of squared distance covariance

    """
    # Initialize the distance covariance matrix
    dcov2_matrix = np.zeros(
        (z_matrix.shape[1], z_matrix.shape[1]),
        dtype=np.float64,
    )

    for j_col in range(z_matrix.shape[1]):
        # Compute the Euclidean distance matrix of the j-th criterion
        j_dmatrix = dist_matrix(z_matrix[:, j_col])

        # Compute the linear function of its Euclidean distance matrix
        j_func = lin_func(j_dmatrix)

        for l_col in range(j_col, z_matrix.shape[1]):
            if j_col == l_col:
                # Compute the distance variance of the j-th criterion
                dcov2_matrix[j_col, j_col] = squared_dcov(j_func, j_func)
            else:
                # Compute the Euclidean distance matrix of the l-th criterion
                l_dmatrix = dist_matrix(z_matrix[:, l_col])

                # Compute the linear function of its Euclidean distance matrix
                l_func = lin_func(l_dmatrix)

                # Compute the squared distance covariance of the two criteria
                dcov2_matrix[j_col, l_col] = squared_dcov(j_func, l_func)
                dcov2_matrix[l_col, j_col] = dcov2_matrix[j_col, l_col]

    return dcov2_matrix


def dist_matrix(z_vector):
    """ Calculate the Euclidean distance matrix of provided vector
    
    Parameters :
    ----------

    z_vector : a pandas or array object that contains specific columns

    Yields:
    -------
    d_matrix : the matrix of the Euclidean distance matrix

    """
    # Initialize the Euclidean distance matrix
    dmatrix = np.zeros(
        (z_vector.shape[0], z_vector.shape[0]),
        dtype=np.float64,
    )

    for i_row in range(z_vector.shape[0]):
        for k_row in range(i_row + 1, z_vector.shape[0]):
            # The Euclidean distance of two real-valued scalars corresponds
            # to the absolute value of their difference
            dmatrix[i_row, k_row] = np.fabs(z_vector[i_row] - z_vector[k_row])
            dmatrix[k_row, i_row] = dmatrix[i_row, k_row]

    return dmatrix


def lin_func(dmatrix):
    """ Calculate the result of the linear function for the provided distance matrix
    
    Parameters :
    ----------

    d_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    res : the result of the linear function for the provided distance matrix
.
    """
    res = dmatrix - np.mean(dmatrix, axis=0) - np.reshape(np.mean(dmatrix, axis=1), (dmatrix.shape[0], 1))+ np.mean(dmatrix)

    return res


def squared_dcov(j_func, l_func):
    """ Calculate the squared distance covariance between the corresponding columns
    
    Parameters :
    ----------
    j_func : an array type object
    l_func : another array type object

    Yields:
    -------
    sq_dcov : the squared distance covariance of j_func and l_func

    """
    sq_dov = np.sum(np.multiply(j_func, l_func)) / (j_func.shape[0] ** 2)
    return sq_dov


def squared_dcor(jl_dcov2, j_dvar2, l_dvar2):

    """ Calculate the squared distance correlation between the corresponding columns
    
    Parameters :
    ----------
    j1_dov2 : an array

    j_dvar2 : an array

    l_dvar2 : an array

    Yields:
    -------
    sq_dcor : the squared distance correlation

    """
    sq_dcor = jl_dcov2 / np.sqrt(j_dvar2 * l_dvar2)
    return sq_dcor



def correlate(z_matrix, c_method):
    """ Calculate the selected correlation coefficients of the provide matrix
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    c_method : the selected correlation method
        - "PEARSON" - Pearson correlation
        - "ABSPEARSON" - Absolute Pearson correlation
        - "DCOR" - Distance correlation

    Yields:
    -------
    the selected correlation cofficients

    """
    # Use the selected correlation method
    if c_method.upper() == "PEARSON":
        return pearson(z_matrix)
    elif c_method.upper() == "ABSPEARSON":
        return abspearson(z_matrix)
    elif c_method.upper() == "DCOR":
        return dcor(z_matrix)
    else:
        raise ValueError("Unknown correlation method ({})".format(c_method))


####***********************    2.2 Critic                     ****************###

def critic(z_matrix, c_method="Pearson"):
    """ Calculate the  weight vector of the provided matrix using the Criteria 
    Importance Through Intercriteria Correlation(CRITIC) method
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    c_method : the selected correlation method
        - "PEARSON" - Pearson correlation
        - "ABSPEARSON" - Absolute Pearson correlation
        - "DCOR" - Distance correlation

    Yields:
    -------
    critic_weights : the weight vector of CRITIC

    """
    # Perform sanity checks
    z_matrix = np.array(z_matrix, dtype=np.float64)

    # Compute the standard deviation of each criterion
    sd_vector = np.std(z_matrix, axis=0, dtype=np.float64)

    # Compute the correlation coefficients between pairs of criteria
    corr_matrix = correlate(z_matrix, c_method)

    # Compute the importance of each criterion
    imp_vector = np.zeros(z_matrix.shape[1], dtype=np.float64)
    for j_col in range(z_matrix.shape[1]):
        tmp_sum = 0.0
        for l_col in range(z_matrix.shape[1]):
            tmp_sum += 1.0 - corr_matrix[j_col, l_col]
        imp_vector[j_col] = sd_vector[j_col] * tmp_sum
    
    # Normalize the importance of each criterion
    critic_weights = imp_vector / np.sum(imp_vector)
    return critic_weights


####***********************           2.3 Entropy                        ****************###

def em(z_matrix):
    """ Calculate the weight vector of the provided decision matrix using the Entropy
    Measure(EM) method
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    em_weights : the weight vector of EM

    """
    # Perform sanity checks
    z_matrix = np.array(z_matrix, dtype=np.float64)

    # Compute the normalization constant
    k_constant = 1.0 / np.log(z_matrix.shape[0])

    # Compute the entropy of each criterion
    e_vector = np.zeros(z_matrix.shape[1], dtype=np.float64)
    for j in range(z_matrix.shape[1]):
        tmp_sum = 0.0
        for i in range(z_matrix.shape[0]):
            if z_matrix[i, j] > 0.0:
                tmp_sum += z_matrix[i, j] * np.log(z_matrix[i, j])
        e_vector[j] = -k_constant * tmp_sum

    # The importance of each criterion corresponds to
    # its normalized degree of divergence
    em_weights = (1.0 - e_vector) / np.sum(1.0 - e_vector)
    return em_weights


####***********************           2.4 Mean Weighting               ****************###
def mw(z_matrix):
    """ Calculate the the weight vector of the provided matrix using the Mean
    Weights(MW) method
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    mw_weights : the weight vector of MW

    """
    # Perform sanity checks
    z_matrix = np.array(z_matrix, dtype=np.float64)

    # Each criterion is considered equally important
    mw_weights = np.full(z_matrix.shape[1], 1.0 / z_matrix.shape[1], dtype=np.float64)
    return mw_weights


####***********************           2.5 Standard Deviation Weighting               ****************###

def sd(z_matrix):
    """ Calculate the the weight vector of the provided matrix using the 
    Standard Deviation(SD) method
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    Yields:
    -------
    sd_weights : the weight vector of SD

    """

    z_matrix = np.array(z_matrix, dtype=np.float64)

    # Compute the standard deviation of each criterion
    sd_vector = np.std(z_matrix, axis=0, dtype=np.float64)

    # The importance of each criterion corresponds to
    # its normalized standard deviation
    sd_weights = sd_vector / np.sum(sd_vector)
    return sd_weights


####******************   2.6  Variability and Interdependencies of Criteria method    ****************###

def vic(z_matrix, c_method="dCor"):
    """ Calculate the the weight vector of the provided matrix using the 
    Variablity and Interdependencies of Criteria(VIC) method.
    
    Parameters :
    ----------

    z_matrix : a pandas or array object that contains specific columns

    c_method : the selected correlation method
        - "DCOR" - Distance correlation (default parameter)
        - "PEARSON" - Pearson correlation
        - "ABSPEARSON" - Absolute Pearson correlation
        

    Yields:
    -------
    vic_weights : the weight vector of VIC

    """

    z_matrix = np.array(z_matrix, dtype=np.float64)
    if c_method is None:
        c_method = "dCor"

    # Compute the standard deviation of each criterion
    sd_vector = np.std(z_matrix, axis=0, dtype=np.float64)

    # Compute the correlation coefficients between pairs of criteria
    corr_matrix = correlate(z_matrix, c_method)

    # Compute the importance of each criterion
    imp_vector = np.zeros(z_matrix.shape[1], dtype=np.float64)
    for j_col in range(z_matrix.shape[1]):
        tmp_sum = 0.0
        for l_col in range(z_matrix.shape[1]):
            tmp_sum += corr_matrix[j_col, l_col]
        imp_vector[j_col] = sd_vector[j_col] / tmp_sum

    # Normalize the importance of each criterion
    return imp_vector / np.sum(imp_vector)


####******************   2.7  Analytic Hierachy Processing    ****************###
def ahp(judge_matrix):
    """ Calculate the weights using Analytic Hierachy Processing(AHP)
    
    Parameters :
    ----------

    judge_matrix : a pair-wise matrix
        
    Yields:
    -------
    ahp_weights : the weight vector of AHP

    """

    judge_matrix = np.array(judge_matrix, dtype=np.float64)
    n =  judge_matrix.shape[0]
    eigenvalues, eigenvectors = np.linalg.eig(judge_matrix)
    value_max = eigenvalues.max()
    value_max_index = eigenvalues.argmax()
    CI = (value_max-n)/(n-1)
    RI_list = [0,0,0,0.58,0.9,1.12,1.24,1.32,1.41,1.45,1.49] 
    if CI/RI_list[n]>0.1:
        return 'Judgement Matrix is NOT consistent'
    else:
        print('The Max Eigenvalue is ',value_max)
        ahp_weights = eigenvectors[:,value_max_index]
        ahp_weights = ahp_weights/ahp_weights.sum()
        return ahp_weights


###########################################################################################
###############################        3 MCDM Models        ###############################
###########################################################################################           
            
####***********************    3.1 Simple Additive Weighting scores     ****************###

def Compute_Weighted_Sum_Score(df,criteria_list,weights_array):
    """Computes weighted sum score for specified columns

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns and scores
    
    columns: a list object that includes the columns to include in the score
    
    weights_array : an array containing the weights for each column
    
    Yields
    ------
    temp_df: a sorted copy of the passed dataframe with the score in a 
    new column named 'Weighted Score (Sum)'

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> import mcdm_functions as mcfunc

    >>> data_dict = {'Product': ['A', 'B', 'C', 'D'],
                 'Product Advantage': [13.1,13.2,12.2,13.2],
                 'Strategic Alignment': [9.8,8.2,10.0,9.6],
                 'Technical Feasibility': [20.0,18.7,18.5,17.1],
                 'Market Attractiveness': [15.5,12.3,13.1,13.1]}
    >>> score_data = pd.DataFrame(data_dict)
    >>> score_data = score_data.set_index('Product')
    >>> print(score_data)
    
            Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                         15.5               13.1                  9.8   
    B                         12.3               13.2                  8.2   
    C                         13.1               12.2                 10.0   
    D                         13.1               13.2                  9.6   

             Technical Feasibility  
    Product                         
    A                         20.0  
    B                         18.7  
    C                         18.5  
    D                         17.1  
    
    >>> criteria = ['Market Attractiveness',
                    'Product Advantage',
                    'Strategic Alignment',
                    'Technical Feasibility']

    >>> criteria_weights = np.array([4,6,7,9])

    >>> temp = mcfunc.Compute_Weighted_Sum_Score(score_data,criteria,criteria_weights)
    >>> print(temp)
    
             Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                         15.5               13.1                  9.8   
    B                         12.3               13.2                  8.2   
    C                         13.1               12.2                 10.0   
    D                         13.1               13.2                  9.6   

             Technical Feasibility  Weighted Score (Sum)  
    Product                                               
    A                         20.0                 389.2  
    B                         18.7                 354.1  
    C                         18.5                 362.1  
    D                         17.1                 352.7  
       
    """   
    temp_df = df.copy()
    temp_weights = weights_array.copy()
    
    temp_df['Weighted Score (Sum)'] = 0
    for i in range(len(criteria_list)):
        current_criteria = criteria_list[i]
        current_weight = temp_weights[i]
        temp_df['Weighted Score (Sum)'] += current_weight*temp_df[current_criteria]
    
    temp_df.sort_values(by = 'Weighted Score (Sum)', ascending=False, inplace=True)
    return temp_df   



####***********************    3.2 Multiplicative Exponential Weighting scores     ****************###         
def Compute_Weighted_Product_Score(df,criteria_list,weights_array):
    """Computes weighted product score for specified columns

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns and scores
    
    columns: a list object that includes the columns to include in the score
    
    weights_array : an array containing the weights for each column
    
    Yields
    ------
    temp_df: a sorted copy of the passed dataframe with the score in a 
    new column named 'Weighted Score (Product)'

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> import mcdm_functions as mcfunc

    >>> data_dict = {'Product': ['A', 'B', 'C', 'D'],
                 'Product Advantage': [13.1,13.2,12.2,13.2],
                 'Strategic Alignment': [9.8,8.2,10.0,9.6],
                 'Technical Feasibility': [20.0,18.7,18.5,17.1],
                 'Market Attractiveness': [15.5,12.3,13.1,13.1]}
    >>> score_data = pd.DataFrame(data_dict)
    >>> score_data = score_data.set_index('Product')
    >>> print(score_data)
    
            Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                         15.5               13.1                  9.8   
    B                         12.3               13.2                  8.2   
    C                         13.1               12.2                 10.0   
    D                         13.1               13.2                  9.6   

             Technical Feasibility  
    Product                         
    A                         20.0  
    B                         18.7  
    C                         18.5  
    D                         17.1  
    
    >>> criteria = ['Market Attractiveness',
                    'Product Advantage',
                    'Strategic Alignment',
                    'Technical Feasibility']

    >>> criteria_weights = np.array([0.1, 0.2, 0.3, 0.4])

    >>> temp = mcfunc.Compute_Weighted_Product_Score(score_data,criteria,criteria_weights)
    >>> print(temp)
    
               Market Attractiveness  Product Advantage  Strategic Alignment  \
    Product                                                                  
    A                         15.5               13.1                  9.8   
    B                         12.3               13.2                  8.2   
    C                         13.1               12.2                 10.0   
    D                         13.1               13.2                  9.6   

             Technical Feasibility  Weighted Score (Product)  
    Product                                                   
    A                         20.0                 14.463295  
    B                         18.7                 13.061291  
    C                         18.5                 13.673125  
    D                         17.1                 13.296022   
       
    """   
    temp_df = df.copy()
    temp_weights = weights_array.copy()
    
    temp_df['Weighted Score (Product)'] = 1
    for i in range(len(criteria_list)):
        current_criteria = criteria_list[i]
        current_weight = temp_weights[i]
        temp_df['Weighted Score (Product)'] *= temp_df[current_criteria]**current_weight
    
    temp_df.sort_values(by = 'Weighted Score (Product)', ascending=False, inplace=True)
    return temp_df   


####****   3.3 Modified Technique for Order Preference by Similarity to Ideal Solution ********### 

def Compute_TOPSIS_Score(df,criteria_list,weights_array):
    '''This function computes a TOPSIS score for suppliers based on the specified criteria and weights
    
    
    """Computes weighted product score for specified columns

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns and scores
    
    columns: a list object that includes the columns to include in the score
    
    weights_array : an array containing the weights for each column
    
    Yields
    ------
    temp_df: a sorted copy of the passed dataframe with the score in a 
    new column named 'TOPSIS Score'
    
    Note
    ----
    This TOPSIS implementation does not perform any standardization of the alernative scores or weights. 
    We do this to allow the user to define and experiment with various methods of regularization.
    
    '''
    
    temp_df = df.copy()
    temp_weights_array = weights_array.copy()
    
    #Step 1
    evaluation_matrix = temp_df[criteria_list].values

    #Step 2
    #squared_evaluation_matrix = evaluation_matrix**2
    #normalized_evaluation_matrix = evaluation_matrix/np.sqrt(np.sum(squared_evaluation_matrix,axis=0))
    normalized_evaluation_matrix = evaluation_matrix

    #Step 3
    #temp_weights_array = temp_weights_array/temp_weights_array.sum()
    weighted_matrix = normalized_evaluation_matrix * temp_weights_array

    #Step 4
    PIS = np.max(weighted_matrix, axis=0)
    NIS = np.min(weighted_matrix, axis=0)

    #Step 5
    intermediate = (weighted_matrix - PIS)**2
    Dev_Best = np.sqrt(intermediate.sum(axis = 1))

    intermediate = (weighted_matrix - NIS)**2
    Dev_Worst = np.sqrt(intermediate.sum(axis = 1))

    #Step 6
    Closeness = Dev_Worst/(Dev_Best+Dev_Worst)

    #Step 7
    temp_df['TOPSIS Score'] = Closeness.tolist()
    temp_df.sort_values(by='TOPSIS Score',ascending=False,inplace=True)

    return temp_df


###########################################################################################
###############################        4 Robustness Test    ###############################
###########################################################################################  


##############################################################################
def Robust_Ranking(df, 
                   columns,
                   perturbations,
                   weights_array,
                   weights_min=0,
                   weights_max= 9999999999,
                   perturbation_range = 0.05,
                   weight_norm_method ='divide_by_sum',
                   score_type = 'weighted_sum',
                   top_values = 5,
                   include_plot=False
                  ):
    """Performs a robustness check for defined ranking methods

    Parameters
    ----------
    df : a pandas DataFrame object that contains the specified columns
    
    columns: a list object that includes the columns to include in the score
    
    weights_array: an array containing the weights for each column
    
    weights_min: The minimum aceptable value for weights
    
    weights_max: The maximum aceptable value for weights
    
    perturbation_range: The perturbation range to consider (expressed as a proportion of the mean value)
    
    weight_norm_method: A string specifying the method to use for normalization of perturbed weights
        - 'divide_by_max' divides all weights by the maximum weight
        - 'divide_by_sum' divides all weights by the sum of the weights
    
    score_type: A string specifying the type of score  calculation
        - 'weighted_sum' uses the weighted sum method
        - 'weighted_product' uses the weighted product method
        - 'topsis'
    
    top_values: specifies the number of alternatives to keep from each ranking (highest ranked scores kept)

    Yields
    ------
    sum_counts: dataframe indicating the proportion of times each alternative appears in top ranking
    """
    
    
    temp_perturbations = perturbations
    df_copy = df.copy()
    temp_columns_list = list(columns)
    temp_weights_array = weights_array.copy()
    
    weight_mean = temp_weights_array.mean()
    perturbation_amount = weight_mean*perturbation_range
            
    alter_list = []
    count_column_name = ''
    proportion_column_name = ''
       
    np.random.seed(42)

    for i in range(temp_perturbations):
        perturbation_vector = np.random.uniform(low = -1.0*perturbation_amount,
                                                high = perturbation_amount,
                                                size=len(weights_array))

        perturbed_weights = temp_weights_array + perturbation_vector
        perturbed_weights = np.maximum(perturbed_weights,weights_min)
        perturbed_weights = np.minimum(perturbed_weights,weights_max)
        
        if weight_norm_method == 'divide_by_sum':
            perturbed_weights = Normalize_Weights(perturbed_weights,'divide_by_sum')
        
        elif weight_norm_method == 'divide_by_max':
            perturbed_weights = Normalize_Weights(perturbed_weights,'divide_by_max')
        
        if score_type == 'weighted_sum':
            count_column_name = 'Weighted Sum Count'
            proportion_column_name = 'Weighted Sum Proportion'
            temp_df = Compute_Weighted_Sum_Score(df_copy,
                                                 temp_columns_list,
                                                 perturbed_weights)
            temp_df = temp_df.nlargest(top_values, columns = 'Weighted Score (Sum)')
            temp_list = temp_df.index.values.tolist()
            alter_list.append(temp_list)
            
        elif score_type == 'weighted_product':
            count_column_name = 'Weighted Product Count'
            proportion_column_name = 'Weighted Product Proportion'
            temp_df = Compute_Weighted_Product_Score(df_copy,
                                                 temp_columns_list,
                                                 perturbed_weights)
            temp_df = temp_df.nlargest(top_values, columns = 'Weighted Score (Product)')
            temp_list = temp_df.index.values.tolist()
            alter_list.append(temp_list)
            
        elif score_type == 'topsis':
            count_column_name = 'TOPSIS Count'
            proportion_column_name = 'TOPSIS Proportion'
            temp_df = Compute_TOPSIS_Score(df_copy,
                                           temp_columns_list,
                                           perturbed_weights)
            temp_df = temp_df.nlargest(top_values, columns = 'TOPSIS Score')
            temp_list = temp_df.index.values.tolist()
            alter_list.append(temp_list)
           
    
    sum_counts = Counter(x for sublist in alter_list for x in sublist)
    sum_counts = pd.DataFrame.from_dict(sum_counts,orient='index').reset_index()
    sum_counts = sum_counts.rename(columns={0:count_column_name})
    sum_counts.sort_values(by=count_column_name, ascending=False,inplace=True)
    sum_counts[proportion_column_name] = sum_counts[count_column_name]/(temp_perturbations)
    
    
    if include_plot is True:
        fig, ax = plt.subplots(1,1,figsize = (8,5))

        sum_counts.plot.bar(y=proportion_column_name,
                            ax = ax,
                            color = 'blue',
                            alpha = 0.5,
                            edgecolor = 'k')

        ax.set_title(proportion_column_name,fontsize = 20)
        ax.set_xlabel('Alternative',fontsize = 20)
        ax.set_ylabel('Proportion',fontsize = 20)
        ax.xaxis.set_tick_params(labelsize=15)
        ax.yaxis.set_tick_params(labelsize=15)

        ax.legend_.remove()
        plt.show()
    
    return sum_counts


# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
"""

from importlib_metadata import entry_points
import numpy as np
from collections import Counter
import random # To randomize the probability of a tree spawning instead of a pond
import matplotlib.pyplot as plt # For plotting our grid of colors
import matplotlib.colors as colors # To map colors onto our grid
import time # To sleep the program to keep final image up longer

import matplotlib.pyplot as plt
import math
plt.rcParams.update({'font.family': 'SimHei', 'mathtext.fontset': 'stix'})

###########################################################################################
###############################     1 Queuing Theory        ###############################
###########################################################################################

####***************************     1.1 Normalization       ****************************###

def queue_p0(c,lam,mu):
    """Calculate the probability when there is no people in queue

    Parameters
    ----------
    c: Number of service desks
    lam: Average rate of customer arrivals
    mu: Average rate of system services
        
    Yields
    ------
    p0: the probability when there is no people in queue

    """

    a1 = sum([1/math.factorial(k)*(lam/mu)**k for k in range(c)])
    a2 = (1/math.factorial(c))*(1/(1-lam/(c*mu))*(lam/mu)**c)
    p0 = 1/(a1+a2)
    return p0


def average_queue(c,lam,mu):

    """Calculate the average number of people in queue

    Parameters
    ----------
    c: Number of service desks
    lam: Average rate of customer arrivals
    mu: Average rate of system services
        
    Yields
    ------
    ave_queue: the average number of people in queue

    """
    p0 = queue_p0(c,lam,mu)
    rou = lam/(c*mu)
    b1 = (c*rou)**c*rou
    b2 = math.factorial(c)*(1-rou)**2
    ave_queue = b1/b2*p0
    
    return ave_queue

###########################################################################################
###############################     2 Forest fire       ###############################
###########################################################################################

####***************************     1.1 Forest fire      ****************************###


# Generate the forest based on tree density, returns the generated forest and the breadth and length of the forest
def GenerateForest(sizeX, sizeY, treeDensity):
    treeCount = 0 # Counts number of trees in the forest to see if density is reached
    forestArea = sizeX * sizeY # Get the area of the forest
    numberOfTrees = int(forestArea * treeDensity) # Get number of trees needed to reach the required forest density

    generatedForest = [[0]*sizeX for i in range(sizeY)] # Create an empty 2D array with the forest's specified dimensions

    # Set up a loop to keep spawning trees randomly until the required tree density has been met
    while (treeCount < numberOfTrees):
        # Loop through a 2D array(list) of elements to spawn either a tree or a pond randomly
        # We start with y axis first since it's the rows of the 2D array
        # Then, we do a for loop within this to loop through the columns of the 2D array representing the forest
        for y in range(sizeY):
            for x in range(sizeX):
                if (random.random() <= treeDensity):  # Note that the random determinant is based on the tree density required (e.g. 0.2 chance of spawning a tree)
                    # If randomnized number between 0-1 falls within the tree density range, we spawn a tree. If not, we spawn a pond.
                    # We set the element of the 2D array by calling the coordinates [row][column] aka [y][x]
                    generatedForest[y][x] = 1 # We use 1 for tree, 0 for pond
                    treeCount += 1
                    
                    # If we exceed the number of trees then we stop spawning (since tree density has been met)
                    if (treeCount >= numberOfTrees):
                        break
    
    # Returns the forest 2D array along with its length and breadth for further use
    # Note that we return an array (list). So to manipulate this, we need to call blah[0] for the 2D array, blah[1] for sizeX etc.
    return([generatedForest, sizeX, sizeY]) 

# This function displays the forest by reading the 2D array and generating a grid of colored squares
# 0 maps to pond (blue), 1 maps to tree (green), 2 maps to spreadable fires (red), 3 maps to ashes (black), 4 maps to newly spread fires (red) (see below comments for more info)
def DisplayForest(forestToDisplay):

    # Clear any previous plots
    plt.clf()

    # Create a color map that maps array values to colors (0: water, 1: tree, 2: fire, 3: ash)
    colorMap = colors.ListedColormap(["royalblue", "green", "red", "black"])
    # Set the bounds of the color map to make it discrete values instead of a gradient
    bounds = [0, 1, 2, 3, 4] # Aka 0 = royalblue, 1 = green and so on. 4 is just an arbitrary filler number.
    norm = colors.BoundaryNorm(bounds, colorMap.N)

    # Create a plot using the 2D forest array as a base and map it to the color map with normalized bounds
    plt.imshow(forestToDisplay[0], cmap = colorMap, norm = norm)
    plt.axis('off') #Remove the axes of the plotted graph so only colored squares are displayed
    plt.pause(0.05) # Delay between each redraw of the image (basically delay between refreshing the display)

# Start forest fire
def StartForestFire(forestToBurn):
    #Start the fire in the middle
    middleX = forestToBurn[1] // 2
    middleY = forestToBurn[2] // 2
    forestToBurn[0][middleY][middleX] = 2 # Set fire to be the value 2, start in middle of 2D array

    burning = True # While fire is still burning, keep looping to start new fires

    # Create a 2D array of same size to store which forest tiles started new fires
    # False = Did not start new fire this loop, True = Started a new fire this loop
    # When all tiles did not start new fires, the loop can end (stop attempting to spread forest fires)
    newFireCheck = [[False]*forestToBurn[1] for i in range(forestToBurn[2])]

    # Start burning algorithm
    while burning:
        # We check each tile in the forest for each loop, loop through the rows, then loop through the columns
        # Recall that forestToBurn[1] = sizeX and forestToBurn[2] = sizeY
        # And recall that sizeY = number of rows, and sizeX = number of columns
        for row in range(0, forestToBurn[2]):
            for column in range(0, forestToBurn[1]):

                newFire = False # First, we assume this tile did not start a new fire this loop

                # If this tile was a fire started in the previous loop, in this loop, it attempts to spread new fires
                if (forestToBurn[0][row][column] == 2):

                    # Check up down left right to see if there's a tree. If it's a tree, spread new fire to it
                    # If the fire on this tile was spread to this tile by another tile on this turn, do NOT spread this fire during this loop iteration
                    # We use 4 for new fire on this turn, only for fires AFTER us. We ignore fires BEFORE us because they will be dealt with on the next loop. 
                    # New fires DO NOT spread on this turn. Only OLD fires (from the previous loop) will spread.

                    # Check up
                    if (row-1 >= 0): # Check if index in range (within the forest area bounds)
                        if (forestToBurn[0][row-1][column] == 1):
                            forestToBurn[0][row-1][column] = 2
                            newFire = True # If we spread a fire, we record that this tile spread a new fire this loop
                            # This means that we cannot end the fire spreading loop yet since we need to check again for new fire spreading possibilities

                    # Check right
                    if (column+1 < forestToBurn[1]): #Check if index in range
                        if (forestToBurn[0][row][column+1] == 1):
                            forestToBurn[0][row][column+1] = 4
                            newFire = True

                    # Check down
                    if (row+1 < forestToBurn[2]): #Check if index in range
                        if (forestToBurn[0][row+1][column] == 1):
                            forestToBurn[0][row+1][column] = 4
                            newFire = True

                    # Check left
                    if (column-1 >= 0): #Check if index in range
                        if (forestToBurn[0][row][column-1] == 1):
                            forestToBurn[0][row][column-1] = 2
                            newFire = True

                    # Since this tile was burning the last turn, this turn, make it ashes
                    forestToBurn[0][row][column] = 3

                # If this fire is a new fire, aka another tile spread to this tile on this loop iteration
                # We change it to a normal fire so the loop after this iteration recognizes it as a spreadable fire
                # This also ensures that we display the tile on fire as red
                elif (forestToBurn[0][row][column] == 4): 
                    forestToBurn[0][row][column] = 2

                # We now store whether this tile spread a new fire on this loop iteration in our fireCheck array.
                # We will store FALSE if this tile did NOT spread a new fire this turn.
                # We will store TRUE if this tile spread a new fire this turn.
                # When ALL tiles stored FALSE, that means there are no more trees that we can spread fires to, and the program ends.
                newFireCheck[row][column] = newFire
        
        # If no tiles started a new fire this loop, no more fire spread. Stop loop.
        if not any(True in col for col in newFireCheck): 
            burning = False

        DisplayForest(forestToBurn) # Displays the forest fire spread as each loop completes

    time.sleep(5) # Keeps the final burnt forest image up for 5 seconds
    return(forestToBurn) # Returns the burnt forest (we don't actually need this)




# -*- coding: utf-8 -*-
"""
Created on April 13,  2022

@author: wang Haihua
"""

from importlib_metadata import entry_points
import time
import pandas as pd
import numpy as np
from collections import Counter
import matplotlib.pyplot as plt
plt.rcParams.update({'font.family': 'SimHei', 'mathtext.fontset': 'stix'})

###########################################################################################
###############################     1 Record        ###############################
###########################################################################################

####***************************     1.1 Process Assessment       ****************************###



def process_assess(merge=False,file_name=''):
    """Record the performance of students based on class and homework

    Parameters
    ----------
    merge    : indicates whether the user wants to merge the current inputs to an existing file(.xlsx)
    file_name: if the user set the merge==True, the function will merge the current record in the file named file_name
        
    Yields
    ------
    A DataFrame object contain the records of students' performance value

    """    
    indicators = ['姓名','日期','课次', '活动主体', '活动类型', '活动内容', '数学内容', '数学', '模型内容', '模型', '编程内容', '编程',
              '写作内容', '写作', '合作', '分析内容', '分析总成就','分析(教师辅助)','建模内容', '建模总成就','建模(教师辅助)',
              '求解内容', '求解总成就','求解(教师辅助)', '解释内容','解释总成就','解释(教师辅助)', '验证内容', '验证总成就','验证(教师辅助)', 
              '汇报内容', '汇报总成就','汇报(教师辅助)']
    description = {'姓名':'请输入学生的 \U0001F466 \033[1m 姓名 \033[0m \U0001F467 \n',
               '日期':'请输入上课的 \U0001F4C6 日期 \U0001F4C6，格式为YY/MM/DD \n',
               '课次':'请输入上课的 \U0001F36D 课次 \U0001F36D，输入阿拉伯数字1-30 \n', 
               '活动主体':'请输入活动主体，1代表-学生，2代表-老师 \n', 
               '活动类型':'请输入活动类型，\n 110-学生课堂案例分析 \U0001F31E \U0001F31C 111-学生课后案例分析  \n 120-学生课堂理论学习 \U0001F31E \U0001F31C 121-学生课后理论学习 \n 130 学生课堂写作   \U0001F31E \U0001F31C 131 学生课后写作 \n', 
               '活动内容':'请输入活动内容的简单描述 \n', 
               '数学内容':'关于\U0001F449 数学 \U0001F448 内容的简单描述 \n', 
               '数学':'数学知识内容的等级(1-初中,3-高中,5-大学) \n', 
               '模型内容':'关于\U0001F449 模型 \U0001F448 内容的简单描述 \n',
               '模型':'模型内容的等级 \n 10-初等,\n 3-常规高等(31评价、32优化、33变化、34预测、35解释、36模拟),\n 50-其他 \n', 
               '编程内容':'关于\U0001F449 编程 \U0001F448 内容的简单描述 \n', 
               '编程':'编程内容的等级 \n 10-基础,\n 3-重要第三方库(31Numpy、32Matplotlib、33Pandas),\n 5-模型第三方库(51Model-Insight、52Pulp、53Scipy、54Sklearn、55Statsmodel、56其他) \n',
               '写作内容':'关于\U0001F449 写作 \U0001F448 内容的简单描述 \n', 
               '写作':'写作内容的等级 \n 1-基础学术论文, \n 3-通用建模论文, \n 5-特定模型论文(51评价、52优化、53变化、54预测、55解释、56模拟、57其他) \n', 
               '合作':'合作人数 \n', 
               '分析内容':'简要描述 \U0001F440 分析内容 \U0001F440 \n', 
               '分析总成就':'分析总成就(0-5) \n',
               '分析(教师辅助)':'教师辅助分析得分(0-5) \U0001F511 \n',
               '建模内容':'简要描述 \U0001F44A 建模内容 \U0001F44A \n', 
               '建模总成就':'建模总成就(0-5) \n', 
               '建模(教师辅助)':'教师辅助建模得分(0-5) \U0001F511 \n',
               '求解内容':'简要描述 \U0001F4D0 求解内容 \U0001F4D0 \n', 
               '求解总成就':'求解总成就(0-5) \n',
               '求解(教师辅助)':'教师辅助求解得分(0-5) \U0001F511 \n',
               '解释内容':'简要描述 \U0001F913 解释内容 \U0001F913 \n',
               '解释总成就':'解释总成就(0-5) \n', 
               '解释(教师辅助)':'教师辅助解释得分(0-5) \U0001F511 \n',
               '验证内容':'简要描述 \U0001F98A 建模内容 \U0001F98A \n', 
               '验证总成就':'验证总成就(0-5) \n', 
               '验证(教师辅助)':'教师辅助解验证得分(0-5) \U0001F511 \n',
               '汇报内容':'简要描述 \U0001F3E8 建模内容 \U0001F3E8 \n', 
               '汇报总成就':'汇报总成就(0-5) \n',
               '汇报(教师辅助)':'教师辅助解汇报得分(0-5) \U0001F511 \n'}
    record_dict = {i:[] for i in indicators}
    print('欢迎进入课程评价记录系统!'+'\U0001F9D0'*3)
    print('Loading...\n',end='')
    print('如果你不希望继续输入，可以在输入框中 输入 exit 即可退出')
    flag_continue = 1
    while flag_continue:
        for i in range(20):
            print('\U0001F680',end='',flush=True)
            time.sleep(0.1)
        print('\n')
        for i in record_dict.keys():
            info = input(description[i])
            if len(info)==0:
                record_dict[i].append(None)
            elif info == 'exit':
                flag_continue = 0
                print('终止本次输入')
                record_dict = {i:[] for i in indicators}
                break
            else:
                record_dict[i].append(info)
        whether_continue = input('是否继续输入下一条记录,1-是，2-否')
        if whether_continue == '1':
            pass
        else:
            flag_continue = 0
    for i in range(7):
        print('\U0001F947',end='',flush=True)
        print('\U0001F948',end='',flush=True)
        print('\U0001F949',end='',flush=True)
        time.sleep(0.1)   
    print('\n')
    print('                    \U0001F44C录入完成\U0001F44C                    ')
    df = pd.DataFrame(record_dict)
    if merge:
        data = pd.read_excel(file_name)
        return data.append(df)


from setuptools import setup, find_packages

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='model-insight',
    version='0.66',
    license='MIT',
    author="Wang Haihua",
    author_email='reformship@gmail.com',
    description='A package for learning and teaching mathematical modeling',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url= "https://reformship.github.io/pages(en)/134modelinsight.html",
    project_urls={
    "Documentation(Chinese)": "https://reformship.github.io/pages/134modelinsight.html",
    "Code": "https://github.com/reformship/model-insight",
    "Issue tracker": "https://github.com/reformship/model-insight/issues",},
    keywords='mathematical modeling',
    install_requires=[
          'numpy','pandas','matplotlib','seaborn','scipy','pulp'
      ],
    include_package_data=False,
    #package_data={'': ['src/model_insight/datasets/*.csv','src/model_insight/datasets/*.xls','src/model_insight/datasets/*.xlsx']},

)
